# -*- coding: utf-8 -*-
"""Account class module for HekrAPI"""
import logging

from typing import Union, Dict, List, Optional, Tuple
from aiohttp import ClientSession
from json import JSONDecodeError, loads

from .device import Device, CloudConnector
from .const import DEFAULT_APPLICATION_ID, DEFAULT_WEBSOCKET_HOST
from .exceptions import AccountUnauthenticatedException, AccountDevicesUpdateFailedException, HekrAPIException

_LOGGER = logging.getLogger(__name__)

class Account:
    """Account class for HekrAPI

    Raises:
        ValueError: Account constructor is not provided possible ways to authenticate
        AccountUnauthenticatedException: Account did not authenticate prior to calling method

    Attributes:
        __username (str): Account username
        __password (str, optional): Account password (not required with token)
        __token (str, optional): Authentication 'Bearer' token (not required with password)
        __devices (dict): Dictionary with 'Device' objects belonging to account
        application_id (str): Application ID (has default value)
    """

    BASE_URL = 'https://user-openapi.hekreu.me'

    def __init__(self, username: Optional[str] = None, password: Optional[str] = None,
                 token: Optional[str] = None,
                 application_id: str = DEFAULT_APPLICATION_ID):
        # @TODO: refactor after enabling authentication support
        if not token and not (username and password):
            raise HekrAPIException("At least one authentication method (token, username/password) must be set up")

        self.application_id = application_id

        self.__username = username
        self.__password = password
        self.__token = token

        self.__connectors = {}

        self.__devices: Dict[str, Device] = {}

    def get_connector(self, connect_host: str, connect_port: int = 186):
        if not self.__token:
            raise AccountUnauthenticatedException(account=self)

        key = (connect_host, connect_port)
        if key not in self.__connectors:
            connector = CloudConnector(
                token=self.__token,
                connect_host=connect_host,
                connect_port=connect_port,
                application_id=self.application_id
            )
            self.__connectors[key] = connector
            return connector
        return self.__connectors[key]

    @property
    def connectors(self) -> Dict[Tuple[str, int], CloudConnector]:
        return self.__connectors

    @property
    def devices(self) -> Dict[str, Device]:
        """Device dictionary accessor

        Returns:
            Dict[str, Device] -- dictionary of devices (device_id => Device object)
        """
        return self.__devices

    def _generate_auth_header(self):
        if not self.__token:
            raise AccountUnauthenticatedException(account=self)
        return {'Authorization': 'Bearer ' + self.__token}

    def authenticate(self):
        """Authenticate account with Hekr"""

    async def get_devices(self) -> Dict[str, dict]:
        auth_header = self._generate_auth_header()
        base_url_devices = self.BASE_URL + '/devices?size={}&page={}'

        request_devices = 20
        current_page = 0

        devices_info = dict()
        async with ClientSession() as session:
            more_devices = True
            while more_devices:
                request_url = base_url_devices.format(
                    request_devices,
                    current_page
                )

                async with session.get(request_url, headers=auth_header) as response:
                    content = await response.read()
                    _LOGGER.debug('Received response (%d) for account %s: %s' % (response.status, self, content))
                    try:
                        response_list = loads(content)
                    except JSONDecodeError:
                        raise AccountDevicesUpdateFailedException(account=self,
                                                                  response=response,
                                                                  reason='Received non-JSON response')

                    if response.status != 200:
                        await session.close()
                        reason = 'Account credentials incorrect' if response.status in (401, 403) \
                            else 'Unknown HTTP error'
                        raise AccountDevicesUpdateFailedException(account=self, response=response, reason=reason)

                    more_devices = (len(response_list) == request_devices)

                    devices_info.update({
                        device_info['devTid']: device_info
                        for device_info in response_list
                    })

        return devices_info

    async def update_devices(self, devices_info: Optional[Dict[str, dict]] = None) -> Dict[str, Device]:
        """
        Get devices, and update attributes if an object already exists, or create new ones based on retrieved info.
        :return: Dictionary with new devices indexed by device ID
        """

        if devices_info is None:
            devices_info = await self.get_devices()

        devices = {}
        for device_id, device_attributes in devices_info:
            if device_id in self.__devices:
                self.__devices[device_id].device_info = device_attributes
            else:
                device = Device(device_id=device_id, control_key=device_attributes['ctrlKey'])
                device.connector = self.get_connector(connect_host=device_attributes['dcInfo']['connectHost'])

                devices[device_id] = device

        if devices:
            self.__devices.update(devices)

        return devices



# -*- coding: utf-8 -*-
"""Device class module for Hekr API"""
from json import dumps, loads

import asyncio
import logging
from aiohttp import ClientSession, WSMsgType, client_exceptions
from typing import Tuple, Union, Optional, Any, TYPE_CHECKING, Dict, List, Set, Callable

from .aioudp import RemoteEndpoint, open_remote_endpoint
from .command import Command
from .const import (
    DEFAULT_APPLICATION_ID,
    DeviceConnectionType,
    DeviceResponseState,
    ACTION_DEVICE_AUTH_RESPONSE, ACTION_CLOUD_AUTH_RESPONSE, ACTION_DEVICE_MESSAGE, ACTION_COMMAND_RESPONSE,
    ACTION_HEARTBEAT_RESPONSE, ACTION_HEARTBEAT_REQUEST, ACTION_DEVICE_AUTH_REQUEST, ACTION_COMMAND_REQUEST,
    ACTION_CLOUD_AUTH_REQUEST, DEFAULT_WEBSOCKET_HOST, DEFAULT_WEBSOCKET_PORT)
from .exceptions import *
from .helpers import sensitive_info_filter
from .types import MessageID, Action, ProcessedResponse, HekrCallback, DeviceResponse, DeviceInfo, \
    AnyCommand, CommandData, DeviceID, DevicesDict

if TYPE_CHECKING:
    from .protocol import Protocol
    from aiohttp.client import _WSRequestContextManager

_LOGGER = logging.getLogger(__name__)


class Listener:
    @staticmethod
    def _get_loop_run_executor(*args):
        loop = asyncio.get_event_loop()
        loop.run_in_executor(None, *args)

    def __init__(self, connector: '_BaseConnector',
                 callback_task_function: callable = None,
                 callback_exec_function: callable = None):
        self.connector = connector
        self._callbacks: Set[HekrCallback] = set()
        self._running: Optional[asyncio.Task] = None

        self._callback_exec_function = self._get_loop_run_executor if callback_exec_function is None \
            else callback_exec_function
        self._callback_task_function = asyncio.create_task if callback_task_function is None \
            else callback_task_function

    def __str__(self):
        return '<Hekr:Listener(' + ('running' if self.is_running else 'stopped') + ', ' + str(self.connector) + ')>'

    def add_callback(self, callback: HekrCallback) -> None:
        self._callbacks.add(callback)

    def del_callback(self, callback: HekrCallback) -> None:
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    @property
    def is_running(self):
        return self._running is not None and not self._running.cancelled()

    async def start_receiving(self):
        _LOGGER.debug('Start receiving on %s' % self)
        try:
            if not self.connector.is_connected:
                await self.connector.open_connection()

            while True:
                response_str = await self.connector.read_response()
                _LOGGER.debug('Listener %s received response %s' % (self, response_str))
                response_processed = self.connector.process_response(response_str)

                message_id, state, action, data, hekr_device = response_processed

                callback_task_function = self._callback_task_function
                callback_exec_function = self._callback_exec_function

                for callback in self._callbacks:
                    # process connector callbacks, if any
                    if asyncio.iscoroutinefunction(callback):
                        callback_task_function(callback(hekr_device, message_id, state, action, data))
                    else:
                        callback_exec_function(callback, hekr_device, message_id, state, action, data)

                if hekr_device is not None:
                    for callback in hekr_device.callbacks:
                        # process device callbacks, if any
                        if asyncio.iscoroutinefunction(callback):
                            callback_task_function(callback(hekr_device, message_id, state, action, data))
                        else:
                            callback_exec_function(callback, hekr_device, message_id, state, action, data)

        except asyncio.CancelledError:
            _LOGGER.debug('Closing listener')
        except Exception as e:
            _LOGGER.exception('Caught exception')
        finally:
            _LOGGER.debug('Finally on listener closing')

    def start(self, create_task_func: Callable = asyncio.create_task):
        _LOGGER.debug('Starting listener %s with factory: %s' % (self, create_task_func))
        self._running = create_task_func(self.start_receiving())

    def stop(self):
        if self.is_running:
            self._running.cancel()
        self._running = None


class _BaseConnector:
    """Base endpoint class for implementing other endpoints"""
    def __init__(self, device: Optional['Device'] = None, application_id: str = DEFAULT_APPLICATION_ID):
        self._last_message_id = 0
        self._message_devices: Dict[int, 'Device'] = dict()
        self._devices = set()
        self._listener = None
        self._application_id = application_id

        if device is not None:
            self.attach_device(device)

    @property
    def last_message_id(self) -> int:
        return self._last_message_id

    # base methods
    def attach_device(self, hekr_device: 'Device') -> None:
        """
        Attach device to connector.
        :param hekr_device: Device to attach
        """
        if hekr_device in self.devices:
            raise Exception('Trying to attach device %s when device with same device ID is attached' % hekr_device)

        self._devices.add(hekr_device)

    def detach_device(self, hekr_device: 'Device') -> None:
        if hekr_device not in self._devices:
            # @TODO: check for object similarity?
            raise Exception('Cannot detach an unattached device (%s)' % hekr_device)

        self._devices.remove(hekr_device)

    def is_attached(self, hekr_device: 'Device') -> bool:
        """Checkes whether device is attached."""
        return hekr_device in self._devices

    @property
    def devices(self) -> DevicesDict:
        """
        Access devices attached to this connector.
        :return: Dictionary of attached devices indexed by device ID.
        :rtype: Dict[str, 'Device']
        """
        return {hekr_device.device_id: hekr_device for hekr_device in self._devices}

    @property
    def listener(self) -> Optional[Listener]:
        """Return bound listener, if exists."""
        return self._listener

    def get_listener(self, listener_factory: Callable[['_BaseConnector'], Listener] = Listener) -> Listener:
        """
        Get existing listener, or create a new one.
        :return:
        """
        if self._listener is None:
            self._listener = listener_factory(self)
        return self._listener

    def _get_request_base(self, action: str, hekr_device: Optional['Device'] = None, message_id: int = None) -> (int, Dict[str, Any]):
        """
        Generate base of the action execution request.
        :param action: Action to execute
        :param hekr_device: (optional) Device to generate request for
        :param message_id: (optional) Specific message ID to use
        :return: Message ID, request payload
        """

        if message_id is None:
            message_id = self._last_message_id + 1

        request_dict = {
            "msgId": message_id,
            "action": action,
        }

        if action != ACTION_HEARTBEAT_REQUEST:
            if not hekr_device:
                raise HekrAPIException('Device not set for a "%s" action request' % action)

            if hekr_device.device_id not in self.devices:
                raise Exception('Device with ID %s is not attached' % hekr_device.device_id)

            request_dict["params"] = {
                "devTid": hekr_device.device_id,
                "ctrlKey": hekr_device.control_key,
            }
            if action.startswith("app"):
                request_dict["params"]["appTid"] = self._application_id

        return message_id, request_dict

    def generate_request(self, action: str, params: dict = None, hekr_device: Optional['Device'] = None,
                         message_id: int = None) -> (int, str):
        """
        Generate request string (JSON format).
        :param action: Action name (read README.md for discovered actions list)
        :param params: (optional) Parameters array for actions
        :param hekr_device: (optional) Device to generate request for
        :param message_id: (optional) Message id (default: last message id for set connection)
        :return: Message ID, request payload
        """
        message_id, request_dict = self._get_request_base(action, hekr_device, message_id)

        if isinstance(params, dict):
            request_dict['params'].update(params)

        self._message_devices[message_id] = hekr_device
        self._last_message_id = message_id

        return message_id, dumps(request_dict)

    def process_response(self, response_str: str) -> ProcessedResponse:
        """
        Handle incoming response packet (decode from existing string).
        :param response_str: Response string
        :return: Response state, response action, response contents, related device
        """
        state = DeviceResponseState.UNKNOWN
        data = loads(response_str)
        action = data.get('action')
        message_id = data.get('msgId')
        response_code = data.get('code', 200)  #@TODO: this is a compatibility-oriented default value
        
        hekr_device = None
        if message_id in self._message_devices:
            hekr_device = self._message_devices[message_id]
            del self._message_devices[message_id]

        elif 'params' in data and 'devTid' in data['params']:
            hekr_device = self.devices.get(data['params']['devTid'])

        if action == ACTION_HEARTBEAT_RESPONSE:
            """Handle heartbeat responses"""
            if response_code == 200:
                _LOGGER.debug('Heartbeat executed successfully on device %s', self)
                state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.error('Heartbeat failed on device %s', self)
                state = DeviceResponseState.FAILURE

        elif action in (ACTION_COMMAND_RESPONSE, ACTION_DEVICE_MESSAGE):
            """Handle responses from command requests"""
            if not hekr_device:
                raise HekrAPIException('Device not attached to connector on response: %s' % data)

            if not hekr_device.protocol:
                raise DeviceProtocolNotSetException(device=self)

            if response_code == 200:
                data = hekr_device.protocol.decode(raw=data['params']['data']['raw'])
                command = data[1]
                _LOGGER.debug('Command executed successfully on device %s'
                              if action == ACTION_COMMAND_RESPONSE else
                              'Received command request for device %s', self)
                state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.debug('Command failed on device %s, raw response: %s', self, data)
                state = DeviceResponseState.FAILURE

        return message_id, state, action, data, hekr_device

    async def send_heartbeat(self, hekr_device: Optional['Device'] = None) -> int:
        """
        Send heartbeat with connector.
        :param hekr_device: (optional) Device to send heartbeat to
        :return: Message ID
        """
        message_id, request_str = self.generate_request(ACTION_HEARTBEAT_REQUEST, hekr_device=hekr_device)
        await self.send_request(request_str)
        return message_id

    # attributes and methods to be overridden wholly by inherent connectors
    connection_type: DeviceConnectionType = NotImplemented
    connection_priority: int = NotImplemented

    @property
    def is_connected(self) -> bool:
        """
        Get connection state of the connector.
        :return: Connection is open
        """
        raise NotImplementedError

    async def open_connection(self) -> None:
        """Open connection with connector."""
        raise NotImplementedError

    async def close_connection(self) -> None:
        """Close connection with connector."""
        raise NotImplementedError

    async def send_request(self, request_str: str) -> None:
        """
        Send request to device with connector.
        :param request_str: Request payload
        """
        raise NotImplementedError

    async def read_response(self) -> str:
        """
        Read response from connector
        :return: Response payload
        """
        raise NotImplementedError


class LocalConnector(_BaseConnector):
    """Endpoint via LAN"""
    connection_type = DeviceConnectionType.LOCAL
    connection_priority = 1000

    def __init__(self, host: str, port: int, hekr_device: Optional['Device'] = None,
                 application_id: str = DEFAULT_APPLICATION_ID) -> None:
        super().__init__(hekr_device, application_id)
        self._endpoint: Optional['RemoteEndpoint'] = None
        self._host = host
        self._port = port

    def __str__(self):
        return '<HekrApi:LocalConnector(' + self._host + ':' + str(self._port) + ')>'

    def attach_device(self, hekr_device: 'Device'):
        if self._devices:
            raise HekrAPIException('Cannot attach more than one device to a local socket')
        self._devices = hekr_device

    def detach_device(self, hekr_device: 'Device'):
        if self._devices is None or hekr_device.device_id != self._devices.device_id:
            raise HekrAPIException('Cannot detach an unattached device (%s)' % hekr_device)
        self._devices = None

    @property
    def devices(self) -> DevicesDict:
        return {self._devices.device_id: self._devices}

    async def open_connection(self) -> None:
        if self.is_connected:
            raise HekrAPIException('Local endpoint already established')
        if not self._devices:
            raise HekrAPIException('Device not set for local endpoint')

        _LOGGER.debug('Opening local endpoint on device %s' % self._devices)
        self._endpoint = await open_remote_endpoint(self._host, self._port)
        _LOGGER.debug('Sending authentication request to %s' % self._devices)
        message_id, request_str = self.generate_request(ACTION_DEVICE_AUTH_REQUEST, hekr_device=self._devices)
        await self.send_request(request_str)
        await self.read_response()

    async def close_connection(self) -> None:
        if self._endpoint is None:
            return

        if self._endpoint.closed:
            self._endpoint = None
            return

        _LOGGER.debug('Closing local endpoint on device %s' % self._devices)
        self._endpoint.close()
        if not self._endpoint.closed:
            raise HekrAPIException('Could not close existing endpoint')

        self._endpoint = None

    async def send_request(self, request_str: str) -> None:
        _LOGGER.debug('Sending request via %s with content: %s' % (self, request_str))
        self._endpoint.send(str.encode(request_str))

    async def read_response(self) -> str:
        _LOGGER.debug('Starting receiving on %s' % self)
        response = await self._endpoint.receive()
        _LOGGER.debug('Received response on %s with content: %s' % (self, response))
        return response.decode('utf-8').strip()

    @property
    def is_connected(self) -> bool:
        return self._endpoint is not None

    def process_response(self, response_str: str) -> ProcessedResponse:
        message_id, state, action, data, device = super().process_response(response_str)
        if state == DeviceResponseState.UNKNOWN:
            response_code = data.get('code')

            if action == ACTION_DEVICE_AUTH_RESPONSE:
                """Handle local authentication responses"""
                if response_code == 200:
                    _LOGGER.debug('Authentication executed successfully on device %s', self)
                    state = DeviceResponseState.SUCCESS
                else:
                    _LOGGER.error('Authentication failed on device %s', self)
                    state = DeviceResponseState.FAILURE

        return message_id, state, action, data, device


class CloudConnector(_BaseConnector):
    """Endpoint via cloud"""
    connection_type = DeviceConnectionType.CLOUD
    connection_priority = 2000

    def __init__(self, token: str, device: Optional['Device'] = None, application_id: str = DEFAULT_APPLICATION_ID,
                 connect_host: str = DEFAULT_WEBSOCKET_HOST, connect_port: int = DEFAULT_WEBSOCKET_PORT) -> None:
        super().__init__(device, application_id)
        self.__token = token
        self._session: Optional[ClientSession] = None
        self._endpoint: Optional['_WSRequestContextManager'] = None
        self._connect_host = connect_host
        self._connect_port = connect_port

    def __str__(self) -> str:
        return '<Hekr:CloudConnector(' + self._connect_host + ':' + str(self._connect_port) + ')>'

    @property
    def is_connected(self) -> bool:
        return self._endpoint is not None and self._session is not None

    async def open_connection(self) -> None:
        _LOGGER.debug('Opening cloud endpoint to device %s' % self._connect_host)
        session = ClientSession()
        try:
            from random import getrandbits
            from base64 import b64encode
            raw_key = bytes(getrandbits(8) for _ in range(16))
            websocket_key = b64encode(raw_key).decode()

            self._endpoint = await session.ws_connect(
                'https://' + self._connect_host + ':' + str(self._connect_port) + '/',
                headers={
                    'Sec-WebSocket-Key': websocket_key,
                    'Sec-WebSocket-Version': '13',
                    'Connection': 'upgrade',
                    'Upgrade': 'websocket'
                })
            self._session = session
            _LOGGER.debug('Cloud endpoint opened on device %s', self)
            _, request_str = self.generate_request(ACTION_CLOUD_AUTH_REQUEST)
            await self.send_request(request_str)
            response_str = await self.read_response()
            self.process_response(response_str)

        except client_exceptions.ClientConnectionError:
            _LOGGER.exception('Client connection could not be established')
            await session.close()

    async def close_connection(self) -> None:
        if not self.is_connected:
            return None

        try:
            await self._endpoint.close()
            await self._session.close()

        except BaseException:
            _LOGGER.exception('Exception occurred while trying to close %s' % self)
            raise

        finally:
            self._session = None
            self._endpoint = None

    def _get_request_base(self, action: str, hekr_device: Optional['Device'] = None, message_id: int = None) -> (int, Dict[str, Any]):
        if action == ACTION_CLOUD_AUTH_REQUEST:
            if message_id is None:
                message_id = self._last_message_id

            request_dict = {
                "msgId": message_id,
                "action": action,
                "params": {
                    "appTid": self._application_id,
                    "token": self.__token
                }
            }

            self._last_message_id = message_id

        else:
            message_id, request_dict = super()._get_request_base(action, hekr_device, message_id)

            if action == ACTION_CLOUD_AUTH_REQUEST:
                if 'params' in request_dict and 'token' not in request_dict['params']:
                    request_dict['params']['token'] = self.__token

        return message_id, request_dict

    async def send_request(self, request_str: str) -> None:
        await self._endpoint.send_str(request_str)

    async def read_response(self) -> str:
        message = await self._endpoint.receive()
        if message.type == WSMsgType.TEXT:
            response_str = message.data
        elif message.type == WSMsgType.BINARY:
            response_str = message.data.decode('utf-8')
        else:
            raise Exception('Unknown response from WebSockets: %s', message)
        return response_str

    def process_response(self, response_str: str) -> ProcessedResponse:
        message_id, state, action, data, device = super().process_response(response_str)
        if state == DeviceResponseState.UNKNOWN:
            response_code = data.get('code')

            if action == ACTION_CLOUD_AUTH_RESPONSE:
                """Handle cloud authentication responses"""
                if response_code == 200:
                    _LOGGER.debug('Cloud authentication executed successfully on device %s', self)
                    state = DeviceResponseState.SUCCESS
                else:
                    _LOGGER.error('Cloud authentication failed on device %s', self)
                    state = DeviceResponseState.FAILURE

        return message_id, state, action, data, device


class Device:
    """Device class for Hekr API"""

    def __init__(self, device_id: DeviceID, control_key: str, protocol: Optional['Protocol'] = None,
                 device_info: Dict[str, Any] = None):
        self.protocol = protocol
        self.device_id = device_id
        self.control_key = control_key
        self._device_info = device_info
        self.heartbeat_interval = 30

        self._connector: Optional[_BaseConnector] = None
        self._callbacks: Set[HekrCallback] = set()

        self.__last_frame_number = 0

    def __str__(self) -> str:
        """
        Generates a string representation of the device
        :return: String representation (human-readable)
        """

        return '<Hekr:Device(' + self.device_id + ')>'

    def __repr__(self) -> str:
        """
        Generates debug string representation of the device
        :return: String representation (python-like)
        """
        return '<{}({}, {})>'.format(
            self.__class__.__name__,
            self.device_id,
            self._connector if self._connector else 'no connector'
        )

    def __eq__(self, other: 'Device') -> bool:
        """
        Compare device ID to other device's ID.
        :param other: Other device
        :return: Comparison result
        """
        if not isinstance(other, Device):
            raise Exception('Comparison with type "%s" is not implemented' % type(other))
        return self.device_id == other.device_id

    def __hash__(self) -> int:
        """
        Generate hash of the device (primarily for sets).
        :return: Hash of the device ID
        """
        return hash(self.device_id)

    @property
    def callbacks(self):
        return self._callbacks

    # callback management
    def add_callback(self, callback: HekrCallback) -> None:
        self._callbacks.add(callback)

    def del_callback(self, callback: HekrCallback) -> None:
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    # connection management
    @property
    def connector(self) -> Optional[_BaseConnector]:
        return self._connector

    @connector.setter
    def connector(self, connector: Optional[_BaseConnector]) -> None:
        old_connector = self._connector
        if old_connector is not None and old_connector.is_attached(self):
            old_connector.detach_device(self)
        self._connector = connector
        if connector is not None:
            connector.attach_device(self)

    async def open_connection(self) -> _BaseConnector:
        """Open connection with available connector."""
        connector = self.connector
        if connector is None:
            raise HekrAPIException('No connector attached to device')

        if not connector.is_connected:
            await connector.open_connection()

        return connector

    async def close_connection(self) -> None:
        """Close open connector for the device."""
        connector = self.connector
        if connector is None:
            raise HekrAPIException('No connector attached to device')

        if connector.is_connected:
            await connector.close_connection()

    # request management
    async def make_request(self, action: Action, params: dict = None,
                           message_id: Optional[MessageID] = None) -> MessageID:
        """
        Make request to device.

        :param action: Action name.
        :param params: Request parameters.
        :param message_id: Message ID.
        :type action: str
        :type params: dict
        :type message_id: int
        :return: Message ID
        """
        connector = await self.open_connection()

        (message_id, request_str) = connector.generate_request(action=action, params=params, hekr_device=self,
                                                               message_id=message_id)

        _LOGGER.debug('Composed request for device %s, content: %s', self, sensitive_info_filter(request_str))

        await connector.send_request(request_str)
        return message_id

    async def get_response(self, message_id: Optional[MessageID] = None) -> DeviceResponse:
        """
        Read response from device and process.
        :param message_id:
        :return:
        """
        connector = await self.open_connection()

        response_str = await connector.read_response()
        resp_message_id, state, action, data, device = connector.process_response(response_str)

        if message_id is not None and message_id != resp_message_id:
            raise HekrAPIException('Received message for a different request (expected mID: %d, got: %d)'
                                   % (message_id, resp_message_id))

        _LOGGER.debug('Received response for device %s: %s', self, sensitive_info_filter(response_str))

        return resp_message_id, state, action, data

    # shorthand request commands
    async def heartbeat(self) -> int:
        """Send heartbeat message

        Keyword Arguments:
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            HeartbeatFailedException: Heartbeat message sending failed
        """
        return await self.make_request(ACTION_HEARTBEAT_REQUEST)

    async def command(self, command: AnyCommand, data: CommandData = None, frame_number: int = None) -> MessageID:
        """
        Execute device command.
        :param command: Command ID/name/object
        :param data: (optional) Data values for datagram
        :param frame_number: (optional) Frame number
        :return: Message ID
        """
        if not isinstance(command, Command):
            if not self.protocol:
                raise DeviceProtocolNotSetException(self)
            command = self.protocol.get_command(command)

        if frame_number is None:
            frame_number = self.__last_frame_number + 1

        self.__last_frame_number = frame_number

        raw = self.protocol.encode(
            data=data,
            command=command,
            frame_number=frame_number
        )

        return await self.make_request(ACTION_COMMAND_REQUEST, {"data": {"raw": raw}})

    # device info-related accessors
    @property
    def device_info(self) -> Optional[DeviceInfo]:
        """
        Accessor to get device info and raise exception if it is not set.
        :return: Device info response from cloud, if set
        """
        return self._device_info

    @device_info.setter
    def device_info(self, new_info: DeviceInfo) -> None:
        """
        Update device info with provided values.
        :param new_info: Device info response from cloud
        """
        new_info = {**new_info}
        if self._device_info is None:
            self._device_info = new_info
        else:
            self._device_info.update(new_info)

        if 'ctrlKey' in new_info:
            self.control_key = new_info['ctrlKey']

    @property
    def _device_info_for_property(self) -> DeviceInfo:
        """
        Shorthand accessor to get device info and raise exception if it is not set.
        :return: Device info response from cloud
        :raises HekrAPIException: No device info is added before property access
        """
        device_info = self._device_info
        if device_info is None:
            raise HekrAPIException('Device info has not been updated because device appears to be created standalone.')
        return self.device_info

    @property
    def product_name(self) -> str:
        return self._device_info_for_property['productName']['en_US']

    @property
    def category_name(self) -> str:
        return self._device_info_for_property['categoryName']['en_US']

    @property
    def sdk_version(self) -> str:
        return self._device_info_for_property['sdkVer']

    @property
    def firmware_version(self) -> str:
        return self._device_info_for_property['binVersion']

    @property
    def url_logo(self) -> str:
        return self._device_info_for_property['logo']

    @property
    def mac_address(self) -> str:
        return self._device_info_for_property['mac']

    @property
    def lan_address(self) -> str:
        return self._device_info_for_property['lanIp']

    @property
    def wan_address(self) -> str:
        return self._device_info_for_property['gis']['ip']['ip']

    @property
    def is_online(self) -> bool:
        return self._device_info_for_property['online']

    @property
    def rssi(self) -> int:
        return self._device_info_for_property['rssi']

    @property
    def bind_key(self) -> str:
        return self._device_info_for_property['bindKey']

    @property
    def device_name(self):
        return self._device_info_for_property['deviceName']



""" Helpers for Hekr API """
from typing import TYPE_CHECKING, Dict, Union, Any, Tuple
from re import sub
from random import getrandbits
from base64 import b64encode

from .const import FRAME_START_IDENTIFICATION
from .exceptions import *

if TYPE_CHECKING:
    from .protocol import Protocol
    from .command import Command




def sensitive_info_filter(content: str) -> str:
    """
    Filters sensitive information from string for logs.

    :param content: Content to filter.
    :type content: str
    :return: Filtered content.
    :rtype: str
    """
    return sub(r'("(ctrlKey|token)"\s*:\s*")[^"]+(")',
               r'\1<redacted>\3',
               content)


def device_id_from_mac_address(mac_address: Union[str, bytearray]) -> str:
    """
    Convert device's physical address to a device ID
    :param mac_address: MAC-address in dash/colon/spaced/concatenated format
    :type mac_address: str, bytes, bytearray
    :return: Device ID
    :rtype: str
    """
    if isinstance(mac_address, (bytes, bytearray)):
        mac_address = mac_address.hex()

    for delimiter in [':', '-', ' ']:
        mac_address = mac_address.replace(delimiter, '')

    mac_address = mac_address.upper()
    return 'ESP_2M_' + mac_address



# -*- coding: utf-8 -*-
"""Command class module for HekrAPI"""

from typing import List

from .types import CommandData
from .argument import Argument
from .exceptions import HekrTypeError, HekrValueError, InvalidDataMissingKeyException, InvalidDataGreaterThanException, \
    InvalidDataLessThanException
from .const import FrameType


class Command:
    """Command class for HekrAPI"""

    def __init__(self,
                 command_id: int,
                 frame_type: FrameType = FrameType.SEND,
                 name: str = None,
                 arguments: list = None,
                 response_command_id: int = None,
                 invoke_command_id: int = None):
        """
        Command class constructor.

        :param command_id: Command identifier for datagrams.
        :param frame_type: Frame type for command.
        :param name: Command name (rarely used externally).
        :param arguments: List of arguments bound to command.
        :param response_command_id: Command ID to wait for response from.
        :param invoke_command_id: Command ID to expect invokation from.
        :type command_id: int
        :type frame_type: FrameType
        :type name: str
        :type arguments: list
        :type response_command_id: int
        """

        # circular checking prevents following private attributes
        # from being created automatically
        self.__response_command_id = None
        self.__invoke_command_id = None
        self.__frame_type = None

        self.command_id = command_id
        self.name = name
        self.frame_type = frame_type
        self.response_command_id = response_command_id
        self.invoke_command_id = invoke_command_id
        self.arguments = arguments

    def __repr__(self) -> str:
        """
        Get pythonic command object representation.

        :return: Command representation (Python-like)
        :rtype: str
        """
        return '<{}({}, {}, {})>'.format(
            self.__class__.__name__,
            '"' + self.name + '"' if self.name is not None else None,
            self.command_id,
            self.frame_type.name
        )

    def __str__(self) -> str:
        """
        Command name alternative getter.

        :return: Command name
        :rtype: str
        """
        return self.name

    def __int__(self) -> int:
        """
        Command ID alternative getter.

        :return:
        """
        return self.command_id

    @property
    def command_id(self) -> int:
        """
        Command ID getter.

        :return: Command ID.
        :rtype: int
        """
        return self.__command_id

    @command_id.setter
    def command_id(self, value: int) -> None:
        """
        Command ID setter.

        :param value: Command ID.
        :raises:
            HekrTypeError: Raised when the value fed is not an integer.
            HekrValueError: Raised when a value is out of bounds of bytes constraint.
        """
        if not isinstance(value, int):
            raise HekrTypeError(variable='command_id',
                                expected=int,
                                got=type(value))
        elif 255 < value < 0:
            raise HekrValueError(variable='command_id',
                                 expected='integer from 0 to 255',
                                 got=value)
        self.__command_id = value

    @property
    def frame_type(self) -> FrameType:
        """
        Getter for `frame_type` attribute.

        :return: Frame type for command.
        :rtype: FrameType
        """
        return self.__frame_type

    @frame_type.setter
    def frame_type(self, value: FrameType) -> None:
        """
        Setter for `frame_type` attribute.

        Coerces any incoming value on invocation to `FrameType` class instances.
        Throws exceptions when incorrect data is being fed.

        :param value: Frame type for command.
        :raises:
            HekrTypeError: Raised when the value is not coercible to a `FrameType` instance
            HekrValueError: Raised when the value is not `FrameType.SEND` when `response_command_id` attribute is set.
            HekrValueError: Raised when the value is not `FrameType.RECEIVE` when `invoke_command_id` attribute is set.
        """
        try:
            if value is None:
                new_frame_type = FrameType.SEND
            elif isinstance(value, str):
                new_frame_type = FrameType[value.upper()]
            else:
                new_frame_type = FrameType(value)
        except KeyError:
            raise HekrValueError(variable='frame_type',
                                 expected=', '.join(map(lambda x: "'"+x.name.lower()+"'", FrameType)) +
                                          " (case-insensitive)",
                                 got=value)
        except ValueError:
            raise HekrTypeError(variable='frame_type',
                                expected='coercible to `%s`' % FrameType.__name__,
                                got=type(value))

        if new_frame_type != FrameType.SEND and self.response_command_id is not None:
            raise HekrValueError(variable='frame_type',
                                 expected='`%s` due to `response_command_id` being set' % FrameType.SEND,
                                 got=new_frame_type)

        if new_frame_type != FrameType.RECEIVE and self.invoke_command_id is not None:
            raise HekrValueError(variable='frame_type',
                                 expected='`%s` due to `invoke_command_id` being set' % FrameType.RECEIVE,
                                 got=new_frame_type)

        self.__frame_type = new_frame_type

    @property
    def arguments(self) -> List[Argument]:
        """
        Arguments list getter.
        :return: List of arguments.
        :rtype: list[Argument]
        """
        return self.__arguments

    @arguments.setter
    def arguments(self, value: List[Argument]) -> None:
        """
        Arguments list setter.

        Throws exceptions when incorrect data is being fed.

        :param value: List of arguments for command.
        :raises:
            HekrTypeError: Raised when one or more values is not of `Argument` type.
            HekrTypeError: Raised when arguments list is fed something other than lists.
        """
        if isinstance(value, list):
            invalid_types = [type(x) for x in value
                if not isinstance(x, Argument)]

            if invalid_types:
                raise HekrTypeError(variable='arguments list',
                                    expected=Argument,
                                    got=invalid_types)
        elif value is not None:
            raise HekrTypeError(variable='arguments list',
                                expected=[list, type(None)],
                                got=type(value))

        self.__arguments = value or []

    @property
    def invoke_command_id(self) -> int:
        """
        Invoke command ID getter.

        :return: int
        """
        return self.__response_command_id

    @invoke_command_id.setter
    def invoke_command_id(self, value):
        if value is not None:
            if self.frame_type is None or self.frame_type != FrameType.RECEIVE:
                raise HekrValueError(variable='invoke_command_id',
                                     expected='`None` due to `frame_type` not being set to `FrameType.RECEIVE` '
                                              '(current value `%s`)' % self.frame_type,
                                     got=value)
            elif not isinstance(value, int):
                raise HekrTypeError(variable='invoke_command_id',
                                    expected=int,
                                    got=type(value))
            elif value < 0:
                raise HekrValueError(variable='invoke_command_id',
                                     expected='>= 0',
                                     got=value)
        self.__invoke_command_id = value

    @property
    def response_command_id(self) -> int:
        """
        Response command ID getter.

        :return: int
        """
        return self.__response_command_id

    @response_command_id.setter
    def response_command_id(self, value):
        if value is not None:
            if self.frame_type is None or self.frame_type != FrameType.SEND:
                raise HekrValueError(variable='response_command_id',
                                     expected='`None` due to `frame_type` not being set to `FrameType.SEND` '
                                              '(current value `%s`)' % self.frame_type,
                                     got=value)
            elif not isinstance(value, int):
                raise HekrTypeError(variable='response_command_id',
                                    expected=int,
                                    got=type(value))
            elif value < 0:
                raise HekrValueError(variable='response_command_id',
                                     expected='>= 0',
                                     got=value)
        self.__response_command_id = value

    def encode(self, data: dict, use_variable_names: bool = False, filter_values: bool = True) -> bytes:
        """Encode arguments into an array of bytes."""
        result = bytes()
        for argument in self.arguments:
            key = argument.variable if use_variable_names else argument.name
            value_input = data.get(key, None)

            if value_input is None:
                raise InvalidDataMissingKeyException(data_key=key)

            if argument.value_min is not None and argument.value_min > value_input:
                raise InvalidDataLessThanException(
                    data_key=key,
                    value=value_input,
                    value_min=argument.value_min)

            if argument.value_max is not None and argument.value_max < value_input:
                raise InvalidDataGreaterThanException(
                    data_key=key,
                    value=value_input,
                    value_max=argument.value_max)

            if filter_values:
                if argument.multiplier:
                    value_input /= argument.multiplier

                value_input = round(value_input)

            # @TODO: make better performing solution
            result += value_input.to_bytes(
                argument.byte_length,
                byteorder='big',
                signed=False)

        return result

    def decode(self, data: bytes, use_variable_names: bool = False, filter_values: bool = True,
               ignore_extra: bool = False) -> CommandData:
        """Decode passed data"""
        result = {}
        current_pos = 0
        data_length = len(data)
        for argument in self.arguments:
            key = argument.variable if use_variable_names else argument.name

            next_pos = current_pos + argument.byte_length
            if next_pos > data_length:
                raise InvalidDataMissingKeyException(data_key=key)

            value_output = int.from_bytes(data[current_pos:next_pos], byteorder='big', signed=False)

            # @TODO: decide on whether clamping/exceptions are required on invalid data

            if filter_values:
                if argument.multiplier is not None:
                    value_output *= argument.multiplier
                    if argument.decimals is not None:
                        value_output = round(value_output, argument.decimals)

                value_output = argument.type_output(value_output)

            result[key] = value_output
            current_pos = next_pos

        if not ignore_extra and current_pos < data_length:
            raise Exception('Provided data is longer than expected for command.')

        return result



# -*- coding: utf-8 -*-
""" Basic protocol definition for a smart power meter """
from enum import IntEnum

from ..protocol import Protocol
from ..command import Command, FrameType
from ..argument import Argument

__all__ = [
    "VoltageWarning",
    "CurrentWarning",
    "PowerSupplyWarning",
    "SwitchState",
    "PROTOCOL"]


class VoltageWarning(IntEnum):
    """ Voltage warning status """
    OK = 0
    OVERVOLTAGE = 1
    UNDERVOLTAGE = 2


class CurrentWarning(IntEnum):
    """ Current warning status """
    OK = 0
    CRITICAL = 1


class PowerSupplyWarning(IntEnum):
    """ Backup power supply warning status """
    OK = 0
    INSUFFICIENT = 1


class SwitchState(IntEnum):
    """ Current state of device """
    ON = 0
    OFF = 1
    FAILURE = 2


TO_FLOAT = (int, float)
TO_BOOL = (int, bool)
TO_STR = (int, str)

PROTOCOL = Protocol(
    Command(0, FrameType.SEND, "queryDev", response_command_id=1),
    Command(1, FrameType.RECEIVE, "reportDev", arguments=[
        Argument("phase_count", int, 1, "type"),
        Argument("switch_state", TO_BOOL, 1, "sw"),
        Argument("total_energy_consumed", TO_FLOAT, 4, "total_Energy", multiplier=0.01),
        Argument("warning_voltage", VoltageWarning, 1, "fault_Over_U"),
        Argument("current_energy_consumption", TO_FLOAT, 3, "total_Active_power", multiplier=0.0001),
        Argument("warning_current", CurrentWarning, 1, "fault_Over_I"),
        Argument("delay_timer", int, 2, "tmCd_M", value_min=0, value_max=1440),
        Argument("delay_enabled", TO_BOOL, 1, "tmCdO_Sw"),
        Argument("warning_battery", PowerSupplyWarning, 1, "fault_SurplusDeficiency")
    ]),
    Command(2, FrameType.SEND, "querySet", response_command_id=8),
    Command(3, FrameType.SEND, "setLimit", arguments=[
        Argument("max_current", TO_FLOAT, 2, "top_I", multiplier=0.01, value_min=0, value_max=120),
        Argument("max_voltage", int, 2, "top_U", value_min=0, value_max=500),
        Argument("min_voltage", int, 2, "low_U", value_min=0, value_max=500),
    ]),
    # 4th command unknown
    Command(5, FrameType.SEND, "clear", arguments=[
        Argument("active_energy_import", TO_FLOAT, 4, "import_Active_Energy", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("active_energy_export", TO_FLOAT, 4, "export_Active_Energy", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("total_energy_consumed", TO_FLOAT, 4, "total_Energy", multiplier=0.01, value_min=0, value_max=999999.99),
    ]),
    Command(6, FrameType.SEND, "queryMeterID", response_command_id=7),
    Command(7, FrameType.RECEIVE, "reportMeterID", arguments=[
        Argument("meter_id_1", TO_STR, 1, "meterID1"),
        Argument("meter_id_2", TO_STR, 1, "meterID2"),
        Argument("meter_id_3", TO_STR, 1, "meterID3"),
        Argument("meter_id_4", TO_STR, 1, "meterID4"),
        Argument("meter_id_5", TO_STR, 1, "meterID5"),
        Argument("meter_id_6", TO_STR, 1, "meterID6"),
    ]),
    Command(8, FrameType.RECEIVE, "reportSet", arguments=[
        Argument("max_current", TO_FLOAT, 2, "top_I", multiplier=0.01, value_min=0, value_max=120),
        Argument("max_voltage", int, 2, "top_U", value_min=0, value_max=500),
        Argument("min_voltage", int, 2, "low_U", value_min=0, value_max=500),
        Argument("option_electricity_purchase", TO_FLOAT, 4, "purchaseElectricity", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("option_electricity_residual", TO_FLOAT, 4, "residualElectricity", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("option_electricity_alarm", TO_FLOAT, 4, "electricityAlarm_set", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("option_prepaid_enabled", TO_BOOL, 1, "prepaidFunctionSw")
    ]),
    Command(9, FrameType.SEND, "setSw", arguments=[
        Argument("switch_state", SwitchState, 1, "sw")
    ]),
    Command(10, FrameType.SEND, "queryData", response_command_id=11),
    Command(11, FrameType.RECEIVE, "reportData", arguments=[
        Argument("current_1", TO_FLOAT, 3, "I1", multiplier=0.001, value_min=0, value_max=999.999),
        Argument("current_2", TO_FLOAT, 3, "I2", multiplier=0.001, value_min=0, value_max=999.999),
        Argument("current_3", TO_FLOAT, 3, "I3", multiplier=0.001, value_min=0, value_max=999.999),
        Argument("voltage_1", TO_FLOAT, 2, "I1", multiplier=0.1, value_min=0, value_max=999.9),
        Argument("voltage_2", TO_FLOAT, 2, "I2", multiplier=0.1, value_min=0, value_max=999.9),
        Argument("voltage_3", TO_FLOAT, 2, "I3", multiplier=0.1, value_min=0, value_max=999.9),
        Argument("total_reactive_power", TO_FLOAT, 3, "total_Reactive_Power", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("reactive_power_1", TO_FLOAT, 3, "reactive_Power1", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("reactive_power_2", TO_FLOAT, 3, "reactive_Power2", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("reactive_power_3", TO_FLOAT, 3, "reactive_Power3", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("total_active_power", TO_FLOAT, 3, "total_Active_Power", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("active_power_1", TO_FLOAT, 3, "active_Power1", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("active_power_2", TO_FLOAT, 3, "active_Power2", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("active_power_3", TO_FLOAT, 3, "active_Power3", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("total_power_factor", TO_FLOAT, 2, "total_Constant", multiplier=0.0001, value_min=0, value_max=199.9999),
        Argument("power_factor_1", TO_FLOAT, 2, "constant1", multiplier=0.001, value_min=0, value_max=9.999),
        Argument("power_factor_2", TO_FLOAT, 2, "constant2", multiplier=0.001, value_min=0, value_max=9.999),
        Argument("power_factor_3", TO_FLOAT, 2, "constant3", multiplier=0.001, value_min=0, value_max=9.999),
        Argument("current_frequency", TO_FLOAT, 2, "rate", multiplier=0.01),
        # @TODO: min, max
        Argument("total_energy_consumed", TO_FLOAT, 4, "total_Energy", multiplier=0.01),
        # @TODO: min, max
        Argument("active_energy_import", TO_FLOAT, 4, "import_Active_Energy", multiplier=0.01),
        # @TODO: min, max
        Argument("active_energy_export", TO_FLOAT, 4, "export_Active_Energy", multiplier=0.01),
    ]),
    Command(12, FrameType.SEND, "setTmCmd", arguments=[
        Argument("delay_timer", int, 2, "tmCd_M"),  # @TODO: min, max
        Argument("delay_enabled", TO_BOOL, 1, "tmCdO_Sw")
    ]),
    Command(13, FrameType.SEND, "SetParameter", arguments=[
        Argument("option_electricity_purchase", TO_FLOAT, 4, "purchaseElectricity", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("option_electricity_alarm", TO_FLOAT, 4, "electricityAlarm_set", multiplier=0.01, value_min=0, value_max=999999.99),
        Argument("option_prepaid_enabled", TO_BOOL, 1, "prepaidFunctionSw")
    ])
)



# -*- coding: utf-8 -*-
# pylint: disable=too-many-locals
"""Global constants and enums for HekrAPI module"""

from enum import Enum


class FrameType(Enum):
    """Datagram frame types (per Hekr documentation)"""
    RECEIVE = 0x01
    SEND = 0x02
    DEVICE = 0xFE
    ERROR = 0xFF


#: Device response states
class DeviceResponseState(Enum):
    """Response state of `make_request`"""
    UNKNOWN = 0
    SUCCESS = 1
    FAILURE = 2
    WAIT_NEXT = 3


class DeviceConnectionType(Enum):
    LOCAL = 1
    CLOUD = 2


#: Default application ID used within requests via cloud and local transports
DEFAULT_APPLICATION_ID = 'hekr_python_connector'

#: First character of every presumably valid raw datagram
FRAME_START_IDENTIFICATION = 0x48

#: Default port to send UDP datagrams to
DEFAULT_DEVICE_PORT = 10000

#: Default amount of attempts a request should retry doing itself
DEFAULT_REQUEST_RETRIES = 2

#: Seconds count before attempting to perform next request iteration
DEFAULT_RETRY_DELAY = 1

#: Default host for controlling devices via websockets
DEFAULT_WEBSOCKET_HOST = 'fra-hub.hekreu.me'
DEFAULT_WEBSOCKET_PORT = 186


ACTION_DEVICE_AUTH_REQUEST = 'appDevAuth'
ACTION_DEVICE_AUTH_RESPONSE = 'appDevAuthResp'
ACTION_CLOUD_AUTH_REQUEST = 'appLogin'
ACTION_CLOUD_AUTH_RESPONSE = 'appLoginResp'
ACTION_HEARTBEAT_REQUEST = 'heartbeat'
ACTION_HEARTBEAT_RESPONSE = 'heartbeatResp'
ACTION_COMMAND_REQUEST = 'appSend'
ACTION_COMMAND_RESPONSE = 'appSendResp'
ACTION_DEVICE_MESSAGE = 'devSend'



"""Provide high-level UDP endpoints for asyncio.

Example:

async def main():

    # Create a local UDP enpoint
    local = await open_local_endpoint('localhost', 8888)

    # Create a remote UDP enpoint, pointing to the first one
    remote = await open_remote_endpoint(*local.address)

    # The remote endpoint sends a datagram
    remote.send(b'Hey Hey, My My')

    # The local endpoint receives the datagram, along with the address
    data, address = await local.receive()

    # This prints: Got 'Hey Hey, My My' from 127.0.0.1 port 8888
    print(f"Got {data!r} from {address[0]} port {address[1]}")
"""

__all__ = ['open_local_endpoint', 'open_remote_endpoint']


# Imports

import asyncio
import warnings


# Datagram protocol

class DatagramEndpointProtocol(asyncio.DatagramProtocol):
    """Datagram protocol for the endpoint high-level interface."""

    def __init__(self, endpoint):
        self._endpoint = endpoint

    # Protocol methods

    def connection_made(self, transport):
        self._endpoint._transport = transport

    def connection_lost(self, exc):
        if exc is not None:  # pragma: no cover
            msg = 'Endpoint lost the connection: {!r}'
            warnings.warn(msg.format(exc))
        self._endpoint.close()

    # Datagram protocol methods

    def datagram_received(self, data, addr):
        self._endpoint.feed_datagram(data, addr)

    def error_received(self, exc):
        msg = 'Endpoint received an error: {!r}'
        warnings.warn(msg.format(exc))


# Enpoint classes

class Endpoint:
    """High-level interface for UDP enpoints.

    Can either be local or remote.
    It is initialized with an optional queue size for the incoming datagrams.
    """

    def __init__(self, queue_size=None):
        if queue_size is None:
            queue_size = 0
        self._queue = asyncio.Queue(queue_size)
        self._closed = False
        self._transport = None

    # Protocol callbacks

    def feed_datagram(self, data, addr):
        try:
            self._queue.put_nowait((data, addr))
        except asyncio.QueueFull:
            warnings.warn('Endpoint queue is full')

    def close(self):
        # Manage flag
        if self._closed:
            return
        self._closed = True
        # Wake up
        if self._queue.empty():
            self.feed_datagram(None, None)
        # Close transport
        if self._transport:
            self._transport.close()

    # User methods

    def send(self, data, addr):
        """Send a datagram to the given address."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.sendto(data, addr)

    async def receive(self):
        """Wait for an incoming datagram and return it with
        the corresponding address.

        This method is a coroutine.
        """
        if self._queue.empty() and self._closed:
            raise IOError("Enpoint is closed")
        data, addr = await self._queue.get()
        if data is None:
            raise IOError("Enpoint is closed")
        return data, addr

    def abort(self):
        """Close the transport immediately."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.abort()
        self.close()

    # Properties

    @property
    def address(self):
        """The endpoint address as a (host, port) tuple."""
        return self._transport.get_extra_info("socket").getsockname()

    @property
    def closed(self):
        """Indicates whether the endpoint is closed or not."""
        return self._closed


class LocalEndpoint(Endpoint):
    """High-level interface for UDP local enpoints.

    It is initialized with an optional queue size for the incoming datagrams.
    """
    pass


class RemoteEndpoint(Endpoint):
    """High-level interface for UDP remote enpoints.

    It is initialized with an optional queue size for the incoming datagrams.
    """

    def send(self, data):
        """Send a datagram to the remote host."""
        super().send(data, None)

    async def receive(self):
        """ Wait for an incoming datagram from the remote host.

        This method is a coroutine.
        """
        data, addr = await super().receive()
        return data


# High-level coroutines

async def open_datagram_endpoint(
        host, port, *, endpoint_factory=Endpoint, remote=False, **kwargs):
    """Open and return a datagram endpoint.

    The default endpoint factory is the Endpoint class.
    The endpoint can be made local or remote using the remote argument.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    loop = asyncio.get_event_loop()
    endpoint = endpoint_factory()
    kwargs['remote_addr' if remote else 'local_addr'] = host, port
    kwargs['protocol_factory'] = lambda: DatagramEndpointProtocol(endpoint)
    await loop.create_datagram_endpoint(**kwargs)
    return endpoint


async def open_local_endpoint(
        host='0.0.0.0', port=0, *, queue_size=None, **kwargs):
    """Open and return a local datagram endpoint.

    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host, port, remote=False,
        endpoint_factory=lambda: LocalEndpoint(queue_size),
        **kwargs)


async def open_remote_endpoint(
        host, port, *, queue_size=None, **kwargs):
    """Open and return a remote datagram endpoint.

    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host, port, remote=True,
        endpoint_factory=lambda: RemoteEndpoint(queue_size),
        **kwargs)


# Testing

try:
    import pytest
    pytestmark = pytest.mark.asyncio
except ImportError:  # pragma: no cover
    pass


async def test_standard_behavior():
    local = await open_local_endpoint()
    remote = await open_remote_endpoint(*local.address)

    remote.send(b'Hey Hey')
    data, address = await local.receive()

    assert data == b'Hey Hey'
    assert address == remote.address

    local.send(b'My My', address)
    data = await remote.receive()
    assert data == b'My My'

    local.abort()
    assert local.closed

    with pytest.warns(UserWarning):
        await asyncio.sleep(1e-3)
        remote.send(b'U there?')
        await asyncio.sleep(1e-3)

    remote.abort()
    assert remote.closed


async def test_closed_endpoint():
    local = await open_local_endpoint()
    future = asyncio.ensure_future(local.receive())
    local.abort()
    assert local.closed

    with pytest.raises(IOError):
        await future

    with pytest.raises(IOError):
        await local.receive()

    with pytest.raises(IOError):
        await local.send(b'test', ('localhost', 8888))

    with pytest.raises(IOError):
        local.abort()


async def test_queue_size():
    local = await open_local_endpoint(queue_size=1)
    remote = await open_remote_endpoint(*local.address)

    remote.send(b'1')
    remote.send(b'2')
    with pytest.warns(UserWarning):
        await asyncio.sleep(1e-3)
        assert await local.receive() == (b'1', remote.address)
    remote.send(b'3')
    assert await local.receive() == (b'3', remote.address)

    remote.send(b'4')
    await asyncio.sleep(1e-3)
    local.abort()
    assert local.closed
    assert await local.receive() == (b'4', remote.address)

    remote.abort()
    assert remote.closed


if __name__ == '__main__':  # pragma: no cover
    pytest.main([__file__])



# -*- coding: utf-8 -*-
# pylint: disable=too-many-instance-attributes,too-many-arguments
"""Argument class module for HekrAPI"""

from typing import Callable, Any, Union, Tuple, Optional
from enum import IntEnum

ArgumentType = Callable[[Any], Any]
class Argument:
    """Argument class for HekrAPI"""

    def __init__(self, name: str,
                 value_type: Union[Tuple[ArgumentType, ArgumentType], ArgumentType]=int,
                 byte_length: int=1,
                 variable: Optional[str]=None,
                 multiplier: Optional[Union[int, float]]=None,
                 decimals: Optional[int]=None,
                 value_min: Optional[Union[int, float]]=None,
                 value_max: Optional[Union[int, float]]=None,
                 ):
        """Argument class constructor

        Arguments:
            name {str} -- Argument name

        Keyword Arguments:
            value_type {Union[Tuple[ArgumentType, ArgumentType], ArgumentType]} --
                Input and output type (may be different if defined as tuple) (default: {int})
            byte_length {int} -- Length of input value in bytes (default: {1})
            variable {Optional[str]} -- Variable name (default: argument's name)
            multiplier {Optional[Union[int, float]]} -- What to multiply input/divide output values by
                (default: {None})
            decimals {Optional[int]} -- Decimals to round input value to (in absence
                of explicit setting, extracted from `multiply` attribute) (default: {None})
            value_min {Optional[Union[int, float]]} -- Minimum value during encoding (default: {None})
            value_max {Optional[Union[int, float]]} -- Maximum value during encoding (default: {None})
        """
        self.name = name
        self.variable = variable
        self.byte_length = byte_length

        if isinstance(value_type, IntEnum):
            self.type_input = int
            self.type_output = value_type

        elif isinstance(value_type, tuple):
            self.type_input = value_type[0]
            self.type_output = value_type[1]

        else:
            self.type_input = value_type
            self.type_output = value_type

        self.value_min = value_min
        self.value_max = value_max
        self.decimals = decimals
        self.multiplier = multiplier

    def __repr__(self):
        """Overloaded argument definiton

        Returns:
            str -- Human-friendly argument definition
        """
        return '<{}({})>'.format(self.__class__.__name__, self.name)

    @property
    def decimals(self) -> int:
        """Getter for decimals count to round result to

        Extracts decimal point from multiplier unless a specific value is set
        via getter method.

        Returns:
            int -- Floating point digits
        """
        if self.__decimals is None:
            decimals = str(self.multiplier)[::-1].find('.')
            return 0 if decimals < 0 else decimals

        return self.__decimals

    @decimals.setter
    def decimals(self, value):
        """Getter for decimals count to round result to"""
        self.__decimals = value

    @property
    def variable(self) -> str:
        """Getter for variable name

        Variable name is used in several encoding/decoding scenarios.
        It defaults to `name` attribute if not set explicitly.

        Returns:
            str -- Variable name
        """
        if self.__variable is None:
            return self.name
        return self.__variable

    @variable.setter
    def variable(self, value):
        """Setter for variable name"""
        self.__variable = value


#!/usr/bin/env python3
from distutils.core import setup
from subprocess import check_output

version = 'v0.1.0'
setup(
    name='hekrapi',
    packages=['hekrapi', 'hekrapi.protocols'],
    version=version[1:],
    license='MIT',
    description='Python Hekr IoT API bindings',
    author='Alexander Ryazanov',
    author_email='alryaz@xavux.com',
    url='https://github.com/alryaz/hekrapi-python',
    download_url='https://github.com/alryaz/hekrapi-python/archive/' + version + '.tar.gz',
    keywords=['Hekr', 'API', 'Wisen', 'Smart Devices', 'IoT'],
    install_requires=[
        'aiohttp',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



"""Types for Hekr API project."""

from typing import Tuple, Dict, Any, Union, Optional, TYPE_CHECKING, Callable

DecodeResult = Tuple['Command', Dict[str, Any], int]
MessageID = int
DeviceID = str
Action = str

ProcessedData = Union[dict, DecodeResult]

DeviceResponse = Tuple[MessageID, 'DeviceResponseState', Action, ProcessedData]
ProcessedResponse = Tuple[MessageID, 'DeviceResponseState', Action, ProcessedData, Optional['Device']]

HekrCallback = Callable[[Optional['Device'], MessageID, 'DeviceResponseState', Action, ProcessedData], Any]

DeviceInfo = Dict[str, Any]
AnyCommand = Union[int, str, 'Command']
CommandData = Optional[Dict[str, Any]]
DevicesDict = Dict[DeviceID, 'Device']



# -*- coding: utf-8 -*-
# pylint: disable=too-many-arguments
"""Protocol class module for Hekr API"""
from json import dumps as json_dumps

from typing import Union, List, Dict
from yaml import dump as yaml_dump, SafeDumper

from .command import Command
from .const import FRAME_START_IDENTIFICATION
from .exceptions import CommandNotFoundException, HekrTypeError, HekrValueError, InvalidMessagePrefixException, \
    InvalidMessageLengthException, InvalidMessageChecksumException, InvalidMessageFrameTypeException
from .types import CommandData, AnyCommand, DecodeResult


def remove_none(obj):
    """
    Remove empty values recursively from an iterable dictionary.

    :param obj:
    :return:
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (remove_none(k), remove_none(v))
            for k, v in obj.items() if k is not None and v is not None
        )
    else:
        return obj


class Protocol:
    """ Protocol definition class """

    def __init__(self, *args):
        """
        Protocol definition constructor.

        :param args: Variable-length of arguments, `Command` objects.
        """
        self.commands = list(args)

    @property
    def commands(self) -> List[Command]:
        """
        Commands list getter.

        :return: List of commands.
        :rtype: list[Command]
        """
        return self.__commands

    @commands.setter
    def commands(self, value: List[Command]) -> None:
        """
        Commands list setter.

        Checks whether the value passed is of `list` type and that all elements of the list
        are of `Command` type.
        :param value: List of commands
        :raises:
            HekrTypeError: Raised when passed value is of a different type than `list`
            HekrTypeError: Raised when passed list contains elements of different type(s) than `list`
        """
        if not isinstance(value, list):
            raise HekrTypeError(variable='commands', expected=list, got=type(value))

        invalid_types = [type(x) for x in value
            if not isinstance(x, Command)]
        if invalid_types:
            raise HekrTypeError(variable='commands', expected=Command, got=invalid_types)

        self.__commands = value

    def __getitem__(self, key: Union[int, str]) -> Command:
        """
        Retrieves command definition by name or identifier via square-bracket accessor.

        :param key: Command identifier (name or ID)
        :type key: int, str
        :return: Command object
        :raises:
            CommandNotFoundException: if command is not found.
        """
        return self.get_command(key)

    def decode(self,
               raw: Union[str, bytes, bytearray],
               use_variable_names=False,
               filter_values=True) -> DecodeResult:
        """
        Decode raw datagram
        :param raw: Raw datagram
        :param use_variable_names: Use variable names as data keys
        :param filter_values: Apply filters to values
        :type: protocol: Protocol
        :type: raw: str, bytes, bytearray
        :type: use_variable_names: bool
        :type: filter_values: bool
        :return: command object, dictionary of values, frame number
        :rtype: (Command, dict[str, Any], int)
        """
        if isinstance(raw, str):
            decoded = bytearray.fromhex(raw)
        elif isinstance(raw, bytes):
            decoded = bytearray(raw)
        elif isinstance(raw, bytearray):
            decoded = raw
        else:
            raise HekrTypeError(variable='raw', expected=[str, bytearray], got=type(raw))

        if decoded[0] != FRAME_START_IDENTIFICATION:
            raise InvalidMessagePrefixException(raw)

        frame_length = decoded[1]
        if frame_length != len(decoded):
            raise InvalidMessageLengthException(raw)

        checksum = decoded[-1]
        current_checksum = sum(decoded[:-1]) % 0x100
        if checksum != current_checksum:
            raise InvalidMessageChecksumException(raw)

        frame_type = decoded[2]
        command_id = decoded[4]
        command = self.get_command_by_id(command_id)
        if frame_type != command.frame_type.value:
            raise InvalidMessageFrameTypeException(raw)

        frame_number = decoded[3]

        data = command.decode(decoded[5:-1], use_variable_names, filter_values)

        return command, data, frame_number

    def encode(self,
               command: AnyCommand,
               data: CommandData = None,
               frame_number: int = 1,
               use_variable_names=False,
               filter_values=True) -> str:
        """
        Encode data into raw datagram.
        :param command: Command object/ID/name
        :param data: Dictionary of values
        :param frame_number: Frame number
        :param use_variable_names: Use variable names as data keys
        :param filter_values: Apply filters to values
        :return: Raw datagram
        :rtype: str
        """
        command = self.get_command(command)

        if not data:
            data = {}

        encoded_command = command.encode(
            data=data,
            use_variable_names=use_variable_names,
            filter_values=filter_values
        )

        raw = bytearray()
        raw.append(FRAME_START_IDENTIFICATION)
        raw.append(command.frame_type.value)
        raw.append(frame_number % 256)
        raw.append(command.command_id)
        raw.extend(encoded_command)
        raw.insert(1, len(raw) + 2)
        raw.append(sum(raw) % 0x100)

        return raw.hex().upper()

    def get_command_by_id(self, command_id: int) -> Command:
        """
        Get command definition object by its ID.

        :param command_id: Command ID
        :return: `Command` object
        :rtype: Command
        :raises:
            CommandNotFoundException: Command definition not found within protocol
        """
        for command in self.__commands:
            if command.command_id == command_id:
                return command

        raise CommandNotFoundException(command_id)

    def get_command_by_name(self, name: str) -> Command:
        """
        Get command definition object by its name.

        :param name: Command name
        :return: `Command` object
        :rtype: Command
        :raises:
            CommandNotFoundException: Command definition not found within protocol
        """
        for command in self.__commands:
            if command.name == name:
                return command

        raise CommandNotFoundException(name)

    def get_command(self, command: AnyCommand) -> Command:
        """
        Get command by its ID/name (or cycle-return `Command` object).

        :param command: Command ID/name, `Command` object
        :type command: int, str, Command
        :return:
        :raises:
            HekrTypeError: Bad value type for `command` argument
        """
        if isinstance(command, Command):
            return command
        if isinstance(command, int):
            return self.get_command_by_id(command)
        if isinstance(command, str):
            return self.get_command_by_name(command)

        raise TypeError(
            "Argument 'command' (type %s) does not evaluate to any supported command type" %
            command)

    def get_definition_dict(self, filter_empty=True) -> Dict:
        """
        Returns protocol definition as Python dictionary.

        :param filter_empty: Filter out empty values
        :return: Protocol definition
        :rtype: dict
        """
        definition = {
            "commands": [
                {
                    "name": command.name,
                    "command_id": command.command_id,
                    "frame_type": command.frame_type.name.lower(),
                    "response_command_id": command.response_command_id,
                    "arguments": [
                        {
                            "name": argument.name,
                            "variable": argument.variable,
                            "byte_length": argument.byte_length,
                            "type_input": argument.type_input.__name__,
                            "type_output": argument.type_output.__name__,
                            "value_min": argument.value_min,
                            "value_max": argument.value_max,
                            "multiplier": argument.multiplier,
                            "decimals": argument.decimals,
                        }
                        for argument in command.arguments
                    ] if command.arguments else None
                }
                for command in self.commands
            ]
        }
        return remove_none(definition) if filter_empty else definition

    def get_definition_serialized(self, output_format: str = 'yaml') -> str:
        """
        Returns protocol definition as serialized dictionary.

        :param output_format: Output format
        :type output_format: str
        :return: Serialized definition dictionary
        :rtype: str
        """
        definition = self.get_definition_dict()
        if output_format == 'yaml':
            output = yaml_dump(definition, Dumper=SafeDumper, default_flow_style=False, sort_keys=False)
        elif output_format == 'json':
            output = json_dumps(definition, ensure_ascii=False)
        else:
            raise HekrValueError(variable='output_format', expected=['yaml', 'json'], got=output_format)

        return output

    def print_definition(self, output_format: str = 'yaml') -> None:
        """
        Prints protocol definition.

        :param output_format: Output format
        :type: output_format: str
        """
        print(self.get_definition_serialized(output_format=output_format))



# -*- coding: utf-8 -*-
"""Exception classes for Hekr API"""


class HekrAPIException(BaseException):
    pass


class HekrAPIFormattedException(HekrAPIException):
    """Base exception class for all HekrAPI exceptions"""
    default_message = 'HekrAPI Exception occured'

    def __init__(self, *args, **kwargs):
        self.arguments = list(args)
        arguments_text = ', '.join([str(a) for a in self.arguments])
        kwargs['reason'] = kwargs.get('reason') or 'unknown'
        # @TODO: remove 'message' key from kwargs
        super().__init__(
            (kwargs.get('message', self.default_message)).format(
                **kwargs) + (' (' + arguments_text + ')' if arguments_text else ''))


class InvalidMessageException(HekrAPIFormattedException):
    """Base exception class for other message-related exceptions"""
    default_message = 'Invalid Hekr message provided'


class InvalidMessagePrefixException(InvalidMessageException):
    """Raised when the first character within the datagram does not equate 0x48"""
    default_message = 'Message prefix is not standard'


class InvalidMessageLengthException(InvalidMessageException):
    """Raised when datagram data length does not equate one provided by the datagram"""
    default_message = 'Message length does not match expected'


class InvalidMessageChecksumException(InvalidMessageException):
    """Raised when the calculated checksum does not equate one provided by the datagram"""
    default_message = 'Message checksum does not match expected'


class InvalidMessageFrameTypeException(InvalidMessageException):
    """Raised when deduced command object frame type does not equate one provided by the datagram"""
    default_message = 'Frame type does not match expected'


class InvalidDataException(HekrAPIFormattedException):
    """Base exception class for other data-related exceptions"""
    default_message = 'Invalid Hekr data dictionary provided'


class InvalidDataMissingKeyException(InvalidDataException):
    """Raised when a specific data key is missing from the provided data dictionary"""
    default_message = "Key `{data_key}` not found in provided data dictionary"


class InvalidDataLessThanException(InvalidDataException):
    default_message = "Key `{data_key}` contains value '{value}' that is less than minimum '{value_min}'"


class InvalidDataGreaterThanException(InvalidDataException):
    default_message = "Key `{data_key}` contains value '{value}' that is greater than maximum '{value_max}'"


class CommandNotFoundException(HekrAPIFormattedException):
    """Raised when a command was not found by value"""
    default_message = 'Could not find command'


class CommandFailedException(HekrAPIFormattedException):
    """Raised when command execution response contains an error"""
    default_message = "Could not execute command '{command}' on '{device.device_id}', reason: {reason}"


class HeartbeatFailedException(HekrAPIFormattedException):
    """Raised when heartbeat command failed to execute"""
    default_message = "Could not execute heartbeat on '{device.device_id}', response: {response}"


class AuthenticationFailedException(HekrAPIFormattedException):
    """Base exception class for other authentication-related exceptions"""
    default_message = "Authentication failed"


class LocalAuthenticationFailedException(AuthenticationFailedException):
    """Raised when local authentication fails"""
    default_message = "Local authentication failed"


class CloudAuthenticationFailedException(AuthenticationFailedException):
    """Raised when cloud authentication fails"""
    default_message = "Cloud authentication failed"


class AccountUnauthenticatedException(HekrAPIFormattedException):
    """Raised when account is not authenticated during method call that requires authentication"""
    default_message = 'Account unauthenticated'


class DeviceProtocolNotSetException(HekrAPIFormattedException):
    """Raised when device does not have a protocol set after creation"""
    default_message = "Device does not have a protocol set"


class DeviceConnectionMissingException(HekrAPIFormattedException):
    """Raised when device call cannot be made due to missing connection parameters"""
    default_message = 'Cannot make requests to device until connection parameters are provided'


class AccountDevicesUpdateFailedException(HekrAPIFormattedException):
    """Raised when devices update function encounters a status code other than OK"""
    default_message = 'Devices update on account failed, reason: {reason}'


class HekrAPIExpectedGotException(HekrAPIException):
    """Base exception for variable-expected-got exceptions"""
    default_message = "For variable {variable} expected {expected}, got {got}"

    def __init__(self, variable, got, expected):
        if not isinstance(expected, list):
            expected = [expected]

        expected = ', '.join([
            ('`' + v.__name__ + '`' if isinstance(v, type)
            else str(v))
            for v in expected
        ])

        if not isinstance(got, list):
            got = [got]

        got = ', '.join([('`' + v.__name__ + '`' if isinstance(v, type) else str(v)) for v in got])

        variable = str(variable)
        variable = variable if ' ' in variable else '`' + variable + '`'

        super().__init__(self.default_message.format(variable=variable, got=got, expected=expected))


class HekrTypeError(HekrAPIExpectedGotException):
    default_message = 'Type for {variable} is invalid (expected {expected}; got {got})'


class HekrValueError(HekrAPIExpectedGotException):
    default_message = 'Value(s) for {variable} is invalid (expected {expected}; got {got})'
